(*
Author:  Alexander Krauss <krauss@in.tum.de> (2009)
Author:  Christian Sternagel <c.sternagel@gmail.com> (2009-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2009-2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Usable_Rules
imports
  "../Rewriting/Tcap"
  "../Rewriting/Trs_Impl"
  "../Orderings/Reduction_Pair"
begin


definition qrewrite :: "bool \<Rightarrow> ('f,'v)terms \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term list"
  where "qrewrite nfs Q R t \<equiv> filter (\<lambda> s. (t,s) \<in> qrstep nfs Q (set R)) (rewrite R t)" 

lemma qrewrite_sound: assumes "t \<in> set (qrewrite nfs Q R s)"
  shows "(s,t) \<in> qrstep nfs Q (set R)" using assms unfolding qrewrite_def by auto

lemma qrewrite_complete:
  assumes SN: "SN_on (qrstep nfs Q (set R)) {s}"
  and qstep: "(s,t) \<in> qrstep nfs Q (set R)"
  and nfs: "nfs \<Longrightarrow> Q \<noteq> {} \<Longrightarrow> wwf_qtrs Q (set R)"
  shows "t \<in> set (qrewrite nfs Q R s)" 
  using qstep
proof 
  fix C \<sigma> l r
  assume Q: "\<forall> u \<lhd> l \<cdot> \<sigma>. u \<in> NF_terms Q" and lr: "(l,r) \<in> set R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>"
    and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  from wwf_qtrs_imp_nfs_False_switch[OF nfs, of nfs] have switch:
    "qrstep nfs Q (set R) = qrstep False Q (set R)" .
  note SN = SN[unfolded switch]
  note qstep = qstep[unfolded switch]
  from only_applicable_rules[OF Q] SN_on_imp_wwf_rule[OF SN s lr Q, unfolded wwf_rule_def] have vars: "vars_term r \<subseteq> vars_term l" by auto
  have step: "(s,t) \<in> rstep_r_p_s (set R) (l,r) (hole_pos C) \<sigma>"
    unfolding rstep_r_p_s_def Let_def s t
    by (simp add: lr)
  have step: "t \<in> set (rewrite R s)"
    by (rule rewrite_complete[OF step vars])
  thus ?thesis using qstep unfolding qrewrite_def switch by auto
qed

locale itrans = 
  fixes nfs :: bool
  and Q :: "('f,'v)terms"
  and R :: "('f,'v)rules"
  and ur :: "('f,'v)trs"
  and us :: "('f \<times> nat)set"
  and cn :: "('f \<times> 'f \<times> nat)"
begin
fun i_trans_check :: "('f,'v)term \<Rightarrow> bool" where 
  "i_trans_check (Fun f ts) = 
    ((f,length ts) \<in> us \<and> \<not> (\<exists> (l,r) \<in> set R - ur. match_tcap_below l (set R) (Fun f ts)))"

declare i_trans_check.simps[simp del]

function i_trans :: "('f,'v)term \<Rightarrow> ('f,'v)term"
where "i_trans (Var x) = (Var x)" 
   |  "i_trans (Fun f ts) = (if SN_on (qrstep nfs Q (set R)) {Fun f ts} then (
         if i_trans_check (Fun f ts) then Fun f (map i_trans ts) 
           else (comb cn) (Fun f (map i_trans ts) # map i_trans (qrewrite nfs Q R (Fun f ts)))
  )  else Fun f ts)"
by (pat_completeness, auto)


termination i_trans
proof -
  have "SN (restrict_SN_supt (qrstep nfs Q (set R)))" 
    by (rule ctxt_closed_imp_SN_restrict_SN_supt[OF ctxt_closed_qrstep])
  hence "wf ((restrict_SN_supt (qrstep nfs Q (set R)))\<inverse>)" (is "wf ?R")  by (simp add: SN_iff_wf)
  show ?thesis
  proof 
    show "wf ?R" by fact
  next
    fix f ts t
    assume "SN_on (qrstep nfs Q (set R)) {Fun f ts}"       
      and "t \<in> set ts"
    hence "Fun f ts \<rhd> t" and "SN_on (qrstep nfs Q (set R)) {Fun f ts}" by auto
    thus "(t, Fun f ts) \<in> ?R" "(t, Fun f ts) \<in> ?R" unfolding restrict_SN_supt_def supt_def restrict_SN_def by auto
  next
    fix f ts t
    assume SN: "SN_on (qrstep nfs Q (set R)) {Fun f ts}"       
      and t: "t \<in> set (qrewrite nfs Q R (Fun f ts))"
    from SN qrewrite_sound[OF t] 
    show "(t,Fun f ts) \<in> ?R" 
      unfolding restrict_SN_supt_def restrict_SN_def by auto
  qed
qed


abbreviation
  i_trans_subst ::
    "('f, 'v) subst \<Rightarrow> ('f, 'v) subst"
where
  "i_trans_subst \<sigma> \<equiv> (\<lambda>x. i_trans (\<sigma> x))"
end

fun ur_closed_term_af :: "('f,'v)term \<Rightarrow> ('f,'v)terms \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f \<times> nat)set \<Rightarrow> 'f af \<Rightarrow> bool"
where "ur_closed_term_af (Var x) _ _ _ _ _ = True"
   |  "ur_closed_term_af (Fun f ts) Q R ur us \<pi> = ((f,length ts) \<in> us \<and> (\<forall> i < length ts. i \<in> \<pi> (f, (length ts)) \<longrightarrow> ur_closed_term_af (ts ! i) Q R ur us \<pi>) \<and> 
  (\<forall> (l,r) \<in> R. match_tcap_below l R (Fun f ts) \<longrightarrow> (l,r) \<in> ur) 
)"

abbreviation ur_closed_af :: "('f,'v)trs \<Rightarrow> ('f \<times> nat)set \<Rightarrow> ('f,'v)terms \<Rightarrow> ('f,'v)trs \<Rightarrow> 'f af \<Rightarrow> bool"
where "ur_closed_af ur us Q R \<pi> \<equiv> 
   (\<forall> (l,r) \<in> ur. ur_closed_term_af r Q R ur us \<pi>)"

notation Ground_Context.equiv_class ("[_]")

context ce_af_redpair
begin

lemma i_transI:
  assumes ur: "ur_closed_term_af t Q (set R) ur us \<pi>"
    and sn: "SN_on (qrstep nfs Q (set R)) {t \<cdot> \<sigma>}"
  shows "(t \<cdot> (itrans.i_trans_subst nfs Q R ur us com \<sigma>), itrans.i_trans nfs Q R ur us com (t \<cdot> \<sigma>)) \<in> NS^*"
proof -
  interpret itrans nfs Q R ur us com .
  from assms show ?thesis 
  proof (induct t)
    case (Fun f ts)  
    hence sClosed: "\<And> i. i < length ts \<and> i \<in> \<pi> (f, (length ts)) \<Longrightarrow> ur_closed_term_af (ts ! i) Q (set R) ur us \<pi>" by auto
    let ?Ruc = "(nfs, Q, R, ur, us, com)"
    from \<open>SN_on (qrstep nfs Q (set R)) {Fun f ts \<cdot> \<sigma>}\<close> have sns: "\<And> s. s \<in> set ts \<Longrightarrow> SN_on (qrstep nfs Q (set R)) {s \<cdot> \<sigma>}"
      by (simp add: SN_imp_SN_arg_gen[OF ctxt_closed_qrstep])
    have "i_trans (Fun f ts \<cdot> \<sigma>)
         = i_trans (Fun f (map (\<lambda> t. t \<cdot> \<sigma>) ts))" (is "?l = _") by auto
    also have "\<dots> = Fun f (map i_trans (map (\<lambda> t. t \<cdot> \<sigma>) ts))"
    proof -
      from \<open>ur_closed_term_af (Fun f ts) Q (set R) ur us \<pi>\<close>
      have inst: "\<forall> (l,r) \<in> set R. match_tcap_below l (set R) (Fun f ts) \<longrightarrow> (l,r) \<in> ur" and us: "(f, length ts) \<in> us" by auto
      have "i_trans_check (Fun f ts \<cdot> \<sigma>)" (is ?check)
      proof (rule ccontr)
        assume "\<not> ?check"
        from this us obtain l r \<delta> where nUsable: "(l,r) \<in> set R - ur" 
          and elem: "l \<cdot> \<delta> \<in> [GCFun f (map (\<lambda> t. tcap (set R) (t \<cdot> \<sigma>)) ts)]" (is "_ \<in> ?hs")
          by (auto simp:i_trans_check.simps Ground_Context.match_def)
        have "\<And> t. [tcap (set R) (t \<cdot> \<sigma>)] \<subseteq> [tcap (set R) t]" by (rule tcap_instance_subset)
        hence "?hs \<subseteq> [GCFun f (map (tcap (set R)) ts)]" (is "_ \<subseteq> [?h]") by auto
        with elem have "l \<cdot> \<delta> \<in> [?h]" by auto
        hence "Ground_Context.match ?h l" unfolding Ground_Context.match_def by auto
        with inst nUsable show False by auto
      qed
      with Fun show ?thesis by auto
    qed
    finally have eq1: "?l = Fun f (map (\<lambda> t. i_trans (t \<cdot> \<sigma>)) ts)" (is "_ = ?r") by simp
    let ?tis = "\<lambda> t. t \<cdot> i_trans_subst \<sigma>"
    let ?its = "\<lambda> t. i_trans (t \<cdot> \<sigma>)"
    have "(Fun f (map ?tis ts), Fun f (map ?its ts)) \<in> NS^*" 
    proof -
      {
        fix i
        assume ilen: "i < length ts" and ipi: "i \<in> \<pi> (f, (length ts))"
        with sns have "ts ! i \<in> set ts" and "SN_on (qrstep nfs Q (set R)) {ts ! i \<cdot> \<sigma>}" by auto
        with Fun sClosed ilen ipi have "(?tis (ts ! i), ?its (ts ! i)) \<in> NS^*" by auto
      } note ind = this
      from wm_NS have cc: "ctxt_closed (NS^*)" by blast
      show ?thesis 
      proof (rule af_steps_imp_orient, rule trans_rtrancl, rule refl_rtrancl, rule cc, force)
        show "\<forall>i<length (map ?tis ts). i \<in> \<pi> (f, (length ts)) \<longrightarrow> (map ?tis ts ! i, map ?its ts ! i) \<in> NS^*"
          using ind by auto
        show "\<forall>bef s t aft. length (map ?tis ts) = Suc (length bef + length aft) \<longrightarrow> length bef \<in> \<pi> (f, (length ts)) \<or> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> NS\<^sup>*"
          using af_compat unfolding af_compatible_def by force
      qed
    qed
    also have "Fun f (map ?tis ts) = Fun f ts \<cdot> i_trans_subst \<sigma>" by auto
    finally show ?case using eq1 by auto
  qed simp
qed
end

lemma (in itrans) i_transII:
  assumes sn: "SN_on (qrstep nfs Q (set R)) {t \<cdot> \<sigma>}" 
  shows "(i_trans (t \<cdot> \<sigma>), t \<cdot> i_trans_subst \<sigma>) \<in> (rstep (ce_trs cn))^*
  \<and> (funas_term t \<subseteq> us \<or> (i_trans (t \<cdot> \<sigma>), t \<cdot> i_trans_subst \<sigma>) \<in> (rstep (ce_trs cn))^+) " 
  (is "(i_trans _,_) \<in> ?rel \<and> (_ \<or> _ \<in> ?Rel)  ")
using sn
proof (induct t)
  case (Var x) thus ?case by auto
next
  case (Fun f ts)
  from \<open>SN_on (qrstep nfs Q (set R)) {Fun f ts \<cdot> \<sigma>}\<close> have sns: "\<And> s. s \<in> set ts \<Longrightarrow> SN_on (qrstep nfs Q (set R)) {s \<cdot> \<sigma>}"
    by (simp add: SN_imp_SN_arg_gen[OF ctxt_closed_qrstep])
  let ?ts = "map (\<lambda> t. i_trans (t \<cdot> \<sigma>)) ts"
  let ?ss = "map (\<lambda> t. t \<cdot> i_trans_subst \<sigma>) ts"
  have len: "length ?ts = length ?ss" and other: "\<forall> i. i< length ?ts \<longrightarrow> (?ts ! i, ?ss ! i) \<in> ?rel" using sns Fun  by auto
  have "(Fun f ?ts, Fun f ?ss) \<in> ?rel"
    by (rule all_ctxt_closedE[of UNIV], insert len other, auto)
  hence nearlyDone: "(Fun f ?ts, Fun f ts \<cdot> i_trans_subst \<sigma>) \<in> ?rel" (is "(_,?ti) \<in> _") by auto
  show ?case
  proof (cases "i_trans_check (Fun f ts \<cdot> \<sigma>)")
    case False
    with Fun have id: "i_trans (Fun f ts \<cdot> \<sigma>) = 
       (comb cn) (Fun f ?ts # map i_trans (qrewrite nfs Q R (Fun f ts \<cdot> \<sigma>)))" (is "?it = (comb cn) ?list") 
      by (simp add: o_def) 
    have "Fun f ?ts \<in> set ?list" by simp
    hence steps: "((comb cn) ?list, Fun f ?ts) \<in> (rstep (ce_trs cn))^+" (is "_ \<in> ?r^+") by (rule ce_trs_sound) 
    from nearlyDone steps have steps: "(i_trans (Fun f ts \<cdot> \<sigma>), ?ti) \<in> ?r^+" unfolding id by auto
    show ?thesis using steps by auto
  next
    case True
    with Fun nearlyDone have part1: "(i_trans (Fun f ts \<cdot> \<sigma>), Fun f ts \<cdot> i_trans_subst \<sigma>) \<in> ?rel"
      by (simp add: o_def) 
    let ?fus = "funas_term (Fun f ts) \<subseteq> us"
    show ?thesis 
    proof (intro conjI, rule part1, unfold disj_commute[of ?fus], rule disjCI)
      assume not: "\<not> ?fus"
      with Fun True have id: "i_trans (Fun f ts \<cdot> \<sigma>) = Fun f ?ts" by simp 
      from True have fus: "(f,length ts) \<in> us" by (simp add: i_trans_check.simps)
      with not have "\<exists> t. t \<in> set ts \<and> \<not> funas_term t \<subseteq> us" by auto
      then obtain t where t: "t \<in> set ts" and tus: "\<not> funas_term t \<subseteq> us" by blast
      from t obtain i where i: "i < length ?ts" and ti: "t = ts ! i" by (simp only: set_conv_nth, auto)
      from Fun(1)[OF t _] sns t ti tus have "(i_trans (ts ! i \<cdot> \<sigma>), ts ! i \<cdot> i_trans_subst \<sigma>) \<in> ?Rel" by auto   
      with i have ind: "(?ts ! i, ?ss ! i) \<in> ?Rel" by auto
      have "(Fun f ?ts, Fun f ?ss) \<in> ?Rel" 
        by (rule rtrancl_trancl_into_trancl[OF len other i ind ctxt_closed_rstep])
      with id show "(i_trans (Fun f ts \<cdot> \<sigma>), Fun f ts \<cdot> i_trans_subst \<sigma>) \<in> ?Rel" by auto
    qed
  qed
qed

context ce_redpair
begin

definition c :: "'a" where "c \<equiv> undefined"
definition d :: "'a" where "d \<equiv> undefined"
definition n :: nat where "n \<equiv> SOME n. \<forall> m. m \<ge> n \<longrightarrow> (\<forall> c d.  ce_trs (c,d,m) \<subseteq> NS)"

lemma NS_ce_compat_n: "m \<ge> n \<Longrightarrow> ce_trs (c,d,m) \<subseteq> NS"
proof -
  assume m: "n \<le> m"
  from mp[OF spec[OF someI_ex[OF NS_ce_compat]] m[unfolded n_def]] show "ce_trs (c,d,m) \<subseteq> NS" by blast
qed

lemma ce_orient: "m \<ge> n \<Longrightarrow> (rstep (ce_trs (c,d,m)))^* \<subseteq> NS^*"
  by (rule rtrancl_mono, rule trs_inclusion, rule wm_NS, rule subst_NS, rule NS_ce_compat_n, simp)

abbreviation mono where "mono m \<equiv> ctxt_closed S \<and> ce_trs (c,d,m) \<subseteq> S"
abbreviation monoNS where "monoNS \<equiv> (NS \<union> S)^*"
abbreviation monoS where "monoS \<equiv> monoNS O S O monoNS"

abbreviation Snwf where "Snwf us \<equiv> S \<union> {lr | lr. \<not> funas_term (fst lr) \<subseteq> us}"

lemma mono_ce: assumes m: "mono m"   
  shows "rstep (ce_trs (c,d,m)) \<subseteq> S"
  by (rule trs_inclusion, insert m, auto simp: subst_S)

definition mode_cond where "mode_cond mmode m \<equiv> mmode \<longrightarrow> mono m"
definition mode_left where "mode_left mmode \<equiv> if mmode then NS \<union> S else NS"
definition mode_NS where "mode_NS mmode \<equiv> if mmode then monoNS else NS^*"

lemma mode_cond_left: "mode_cond mmode m \<Longrightarrow> ctxt_closed (mode_left mmode) \<and> subst_closed (mode_left mmode)"
  by (cases mmode,
    auto simp: mode_cond_def mode_left_def subst_S subst_NS wm_NS intro: subst_closed_union mono_NSS)

lemma mode_left_NS: "mode_left mode \<subseteq> mode_NS mode"
  unfolding mode_left_def mode_NS_def
  by (cases mode, auto)

lemma mode_NS_mode[simp]: "NS^* O mode_NS mode = mode_NS mode"
proof (cases mode)
  case False thus ?thesis unfolding mode_left_def mode_NS_def by auto
next
  case True 
  hence id: "mode = True" by simp
  show ?thesis unfolding mode_left_def mode_NS_def id
    by (simp, regexp)
qed

lemma mode_mode_NS[simp]: "mode_NS mode O NS^* = mode_NS mode"
proof (cases mode)
  case False thus ?thesis unfolding mode_left_def mode_NS_def by auto
next
  case True 
  hence id: "mode = True" by simp
  show ?thesis unfolding mode_left_def mode_NS_def id
    by (simp, regexp)
qed
end

lemma (in ce_af_redpair) i_transIII: assumes 
  nfs: "nfs = False" and (* here, instead it would suffice to demand that lhss of R are not variables *)
  mode: "mode_cond mode m" and
  rule: "lr \<in> set R" and
  m: "m \<ge> n" and
  sn: "SN_on (qrstep nfs Q (set R)) {Fun f ts}" and
  step: "(Fun f ts,s) \<in> rqrstep nfs Q {lr}" and
  ur: "itrans.i_trans_check R ur us (Fun f ts)" and
  ur_closed: "ur_closed_af ur us Q (set R) \<pi>" and
  orient: "ur \<subseteq> mode_left mode"
  shows "((itrans.i_trans nfs Q R ur us (c,d,m) (Fun f ts), itrans.i_trans nfs Q R ur us (c,d,m) s) \<in> mode_NS mode) \<and> 
    (mode \<and> lr \<in> Snwf us \<longrightarrow>  (itrans.i_trans nfs Q R ur us (c,d,m) (Fun f ts), itrans.i_trans nfs Q R ur us (c,d,m) s) \<in> monoS) \<and> lr \<in> ur"
proof -
  let ?cn = "(c,d,m)"
  interpret itrans nfs Q R ur us ?cn .
  let ?I = "i_trans"
  obtain l r where lr: "lr = (l,r)" by (cases lr, auto)
  note rule = rule[unfolded lr]
  from step[unfolded lr rqrstep_def] obtain \<sigma> where left: "l \<cdot> \<sigma> = Fun f ts" and right: "r \<cdot> \<sigma> = s" by auto
  let ?Is = "i_trans_subst \<sigma>"
  {
    assume "is_Var l"
    with rule obtain x where "(Var x, r) \<in> set R" by (cases l, auto)
    hence "\<not> (SN_on (qrstep nfs Q (set R)) {Fun f ts})" 
      by (rule left_Var_imp_not_SN_qrstep, insert nfs, auto)
    hence False using sn by auto
  }
  then obtain g ss where l: "l = Fun g ss" by (cases l, auto)
  from ur have nmatch: "\<forall> (l,r) \<in> set R - ur. \<not> match_tcap_below l (set R) (Fun f ts)" by (simp only: i_trans_check.simps, blast)
  from l left have fg: "f = g" and tsss: "ts = map (\<lambda> s. s \<cdot> \<sigma>) ss" (is "_ = ?ts") and nvar: "\<not> (is_Var l)" by auto
  have "\<forall> i. ss ! i \<cdot> \<sigma> \<in> [tcap (set R) (ss ! i \<cdot> \<sigma>)]"
  proof 
    fix i
    show "ss ! i \<cdot> \<sigma> \<in> [tcap (set R) (ss ! i \<cdot> \<sigma>)]"
      by (rule tcap_sound[where s = "ss ! i \<cdot> \<sigma>" and t = "ss ! i \<cdot> \<sigma>" and \<sigma> = Var], auto)
  qed
  hence "match_tcap_below (Fun f ss) (set R) (Fun f (map (\<lambda> s. s \<cdot> \<sigma>) ss))"
    by (simp add: Ground_Context.match_def, blast)
  with l fg tsss rule nmatch have urule: "(l, r) \<in> ur" by blast
  with left and right 
    have urStep: "(Fun f ts, s) \<in> rstep ur" by (auto) (metis rstep.id rstep.subst)
  from urule ur_closed have fr: "ur_closed_term_af r Q (set R) ur us \<pi>" by auto
  from left sn have snl: "SN_on(qrstep nfs Q (set R)) {l \<cdot> \<sigma>}" by auto
  from i_transII[OF snl] left have oneB: "funas_term (fst lr) \<subseteq> us \<or> (?I (Fun f ts), l \<cdot> ?Is) \<in> (rstep (ce_trs ?cn))^+" unfolding lr by auto
  from i_transII[OF snl] ce_orient[OF m] have "(?I (l \<cdot> \<sigma>), l \<cdot> ?Is) \<in> NS^*" by blast
  with left have one: "(?I (Fun f ts), l \<cdot> ?Is) \<in> NS^*" by auto
  from step rule lr have "(Fun f ts, s) \<in> rqrstep nfs Q (set R)" unfolding rqrstep_def by auto
  hence "(Fun f ts, s) \<in> qrstep nfs Q (set R)" unfolding qrstep_iff_rqrstep_or_nrqrstep  by auto
  with right sn step_preserves_SN_on[of _ _ "qrstep nfs Q (set R)"] have sns: "SN_on (qrstep nfs Q (set R)) {r \<cdot> \<sigma>}" by auto
  have "(r \<cdot> ?Is , ?I (r \<cdot> \<sigma>)) \<in> NS^*"  by (rule i_transI[OF fr sns])
  with right have stepRight: "(r \<cdot> ?Is, ?I s) \<in> NS^*" by auto
  from urule have urstep: "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> rstep ur" by auto
  with trs_inclusion[OF _ _ orient] mode_cond_left[OF mode] have lstep: "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> mode_left mode" by auto
  with mode_left_NS have rstep: "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> mode_NS mode" by auto
  with stepRight have two: "(l \<cdot> ?Is, ?I s) \<in> mode_NS mode O NS^*" by blast
  from one two have "(?I (Fun f ts), ?I s) \<in> NS^* O mode_NS mode O NS^*" by blast
  hence part1: "(?I (Fun f ts), ?I s) \<in> mode_NS mode" by simp
  show ?thesis
  proof (intro conjI impI, rule part1, unfold lr, erule conjE)
    show "(l,r) \<in> ur" by (rule urule)
  next
    assume mono: mode and S: "(l,r) \<in> Snwf us"
    from S have "(l,r) \<in> S \<or> \<not> funas_term l \<subseteq> us" by auto
    hence "(?I (Fun f ts), r \<cdot> ?Is) \<in> monoS"
    proof
      assume "(l,r) \<in> S"
      with subst_S have "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> S" unfolding subst_closed_def by auto
      moreover have "(?I (Fun f ts), l \<cdot> ?Is) \<in> monoNS" using one rtrancl_mono by blast
      ultimately show ?thesis by auto
    next
      assume "\<not> funas_term l \<subseteq> us"
      with oneB have steps: "(?I (Fun f ts), l \<cdot> ?Is) \<in> (rstep (ce_trs ?cn))^+" unfolding lr by auto
      from mode[unfolded mode_cond_def] \<open>mode\<close> 
        trancl_mono[OF steps mono_ce]
      have "(?I (Fun f ts), l \<cdot> ?Is) \<in> S^+" by auto
      moreover have "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> CS (NS \<union> S)" using S urule orient \<open>mode\<close> unfolding mode_left_def by auto
      hence "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> NS \<union> S" using subst_NS subst_S unfolding cs_union_distr subst_closed_def by force
      ultimately have "(?I (Fun f ts), r \<cdot> ?Is) \<in> S^+ O (NS \<union> S)" by auto
      moreover have "S^+ O (NS \<union> S) \<subseteq> monoS" by regexp
      ultimately show ?thesis by blast
    qed
    with stepRight have "(?I (Fun f ts), ?I s) \<in> monoS O NS^*" by blast
    moreover have "monoS O NS^* \<subseteq> monoS" by regexp
    ultimately show "(?I (Fun f ts), ?I s) \<in> monoS" by blast
  qed
qed


lemma (in itrans) i_transIV: assumes 
  nfs: "nfs = False" and (* required for completeness of rewrite list in qrewrite *)
  sn: "SN_on (qrstep nfs Q (set R)) {Fun f ts}" and
  step: "(Fun f ts,s) \<in> qrstep nfs Q (set R)" and
  ur: "\<not> i_trans_check (Fun f ts)" 
  shows "(i_trans (Fun f ts), i_trans s) \<in> (rstep (ce_trs cn))^+" (is "_ \<in> ?rel")
proof -
  let ?I = "i_trans"
  from ur sn have id: "?I (Fun f ts) = (comb cn) (Fun f (map ?I ts) # map ?I (qrewrite nfs Q R (Fun f ts)))" (is "_ = (comb cn) ?list") 
    by auto
  from qrewrite_complete[OF sn step] nfs have "s \<in> set (qrewrite nfs Q R (Fun f ts))" by simp
  hence "(?I s) \<in> set ?list" by auto
  hence "(comb cn ?list, ?I s) \<in> ?rel" by (rule ce_trs_sound)
  with id show ?thesis by simp
qed

context ce_redpair
begin

lemma mode_NS: "NS^* \<subseteq> mode_NS mode"
  unfolding mode_NS_def
  by (cases mode, insert rtrancl_mono[of NS], auto)

lemma mode_NS_ctxt: "mode_cond mode m \<Longrightarrow> ctxt_closed (mode_NS mode)"
proof (cases mode)
  case False thus ?thesis unfolding mode_NS_def using wm_NS by auto 
next
  case True
  assume "mode_cond mode m"
  hence "ctxt_closed S" unfolding mode_cond_def using True by auto
  from mono_NSS[OF this] show ?thesis unfolding mode_NS_def using True by auto 
qed

lemma mode_S_ctxt: "mode \<Longrightarrow> mode_cond mode m \<Longrightarrow> ctxt_closed monoS"
proof -
  assume m: mode and c: "mode_cond mode m"
  from mode_NS_ctxt[OF c] have NS: "ctxt_closed monoNS" unfolding mode_NS_def
    using m by auto
  from c m have S: "ctxt_closed S" unfolding mode_cond_def by auto
  from S NS show ?thesis by blast
qed

lemma mode_ce: "mode \<Longrightarrow> mode_cond mode m \<Longrightarrow> (rstep (ce_trs (c,d,m)))^+ \<subseteq> monoS"
proof -
  assume mode and mode: "mode_cond mode m"
  with mode have "mono m" unfolding mode_cond_def by auto
  from trancl_mono[OF _ mono_ce[OF this]] have "(rstep (ce_trs (c,d,m)))^+ \<subseteq> S^+" by auto
  moreover have "S^+ \<subseteq> monoS" by regexp
  ultimately show ?thesis by auto
qed
end

lemma (in ce_af_redpair) i_transV: 
  assumes 
  nfs: "nfs = False" and (* mainly used for i_transIII and i_transIV *)
  mode: "mode_cond mode m" and
  m: "m \<ge> n" and
  rule: "lr \<in> set R" and
  sn: "SN_on (qrstep nfs Q (set R)) {t}" and
  step: "(t,s) \<in> qrstep nfs Q {lr}" and
  ur_closed: "ur_closed_af ur us Q (set R) \<pi>" and
  orient: "ur \<subseteq> mode_left mode" 
  shows "(itrans.i_trans nfs Q R ur us (c,d,m) t, itrans.i_trans nfs Q R ur us (c,d,m) s) \<in> mode_NS mode \<and> 
  (mode \<and> (lr \<notin> ur \<or> lr \<in> Snwf us) \<longrightarrow> (itrans.i_trans nfs Q R ur us (c,d,m) t, itrans.i_trans nfs Q R ur us (c,d,m) s) \<in> monoS)"
proof -
  let ?L = "mode_left mode"
  let ?R = "mode_NS mode"
  let ?cn = "(c,d,m)"
  interpret itrans nfs Q R ur us ?cn .
  let ?I = "i_trans"
  let ?cond = "mode \<and> (lr \<notin> ur \<or> lr \<in> Snwf us)"
  let ?goal = "\<lambda> t s. 
    (?I t, ?I s) \<in> ?R \<and> 
  (?cond \<longrightarrow> (?I t, ?I s) \<in> monoS)"
  from step sn show "?goal t s"
  proof induct 
    case (IH C \<sigma> l r)
    with rule have lr: "lr = (l,r)" and IH2: "(l,r) \<in> set R" by auto
    from IH(4)
    show ?case
    proof (induct C)
      case Hole
      let ?ls = "\<box>\<langle>l \<cdot> \<sigma>\<rangle>" 
      let ?rs = "\<box>\<langle>r \<cdot> \<sigma>\<rangle>" 
      from IH(1) IH(2)
      have step: "(?ls, ?rs) \<in> qrstep nfs Q (set R)" using rule
        by (intro qrstepI[where C = Hole], insert IH(3), auto)
      {
        assume "is_Var ?ls"
        then obtain x where "?ls = Var x" by (cases ?ls, auto)
        then obtain x where l: "l = Var x" by (cases l, auto)
        from left_Var_imp_not_SN_qrstep[OF IH2[unfolded l], of False] nfs IH(4)
        have False by simp
      }
      hence "\<not> (is_Var ?ls)" ..
      from this obtain f ls where ls: "?ls = Fun f ls" by (cases ?ls, auto)
      from IH(3) Hole ls have snls: "SN_on (qrstep nfs Q (set R)) {Fun f ls}" by auto
      from step ls have stepls: "(Fun f ls, ?rs) \<in> qrstep nfs Q (set R)" by auto
      from IH(1,3) have steppls: "(Fun f ls, ?rs) \<in> rqrstep nfs Q {(l,r)}" unfolding ls[symmetric] rqrstep_def by auto
      show "?goal ?ls ?rs"
      proof (cases "i_trans_check (Fun f ls)")
        case True
        from i_transIII[OF nfs mode IH2 m snls steppls True ur_closed orient]
        show ?thesis using IH2 unfolding ls lr by blast
      next
      case False
        from nfs snls stepls this 
        have steps: "(?I ?ls, ?I ?rs) \<in> (rstep (ce_trs ?cn))^+" unfolding ls by (rule i_transIV)
        from steps ce_orient[OF m] have "(?I ?ls, ?I ?rs) \<in> NS^*" by auto
        hence piece0:"(?I ?ls, ?I ?rs) \<in> mode_NS mode" using mode_NS by blast
        show ?thesis 
          by (intro conjI[OF piece0] impI, elim conjE, insert steps mode_ce[OF _ mode], auto)
      qed
    next
      case (More f bef D aft)
      obtain t s where t: "t = l \<cdot> \<sigma>" and s: "s = r \<cdot> \<sigma>" by auto
      from More
      have snbta: "SN_on (qrstep nfs Q (set R)) {Fun f (bef @ D\<langle>t\<rangle> # aft)}" (is "SN_on _ {Fun f ?bta}") 
        unfolding t by auto
      let ?Dt = "D\<langle>t\<rangle>"
      let ?Ds = "D\<langle>s\<rangle>"
      let ?bsa = "bef @ ?Ds # aft"
      let ?C = "More f bef D aft"
      have step: "(Fun f ?bta, Fun f ?bsa) \<in> qrstep nfs Q (set R)" 
        unfolding s t
        by (intro qrstepI[OF IH(1) IH2, where C = ?C], insert IH(3), auto)
      show ?case 
      proof (cases "i_trans_check (Fun f ?bta)")
        case False
        from nfs snbta step False 
        have steps: "(?I (Fun f ?bta), ?I (Fun f ?bsa)) \<in> (rstep (ce_trs ?cn))^+" 
          by (rule i_transIV)
        from steps ce_orient[OF m] have "(?I (Fun f ?bta), ?I (Fun f ?bsa)) \<in> NS^*" by auto
        hence piece0: "(?I (Fun f ?bta), ?I (Fun f ?bsa)) \<in> mode_NS mode" using mode_NS by blast
        show ?thesis unfolding s[symmetric] t[symmetric]
          by (rule conjI, insert piece0, simp, intro impI, elim conjE, insert steps mode_ce[OF _ mode], auto)
      next
        case True
        hence fus: "(f, Suc (length bef + length aft)) \<in> us" by (simp add: i_trans_check.simps)
        have "?Dt \<in> set ?bta" by auto 
        with snbta have "SN_on (qrstep nfs Q (set R)) {?Dt}" by (rule SN_imp_SN_arg_gen[OF ctxt_closed_qrstep])
        with More have "?goal ?Dt ?Ds" unfolding s t by auto
        hence always: "(?I ?Dt, ?I ?Ds) \<in> ?R" and choice: "?cond \<longrightarrow> (?I ?Dt, ?I ?Ds) \<in> monoS" by auto
        from ctxt_closed_one[OF mode_NS_ctxt[OF mode] always]
        have piece0: "(Fun f (map ?I bef @ ?I ?Dt # map ?I aft), 
          Fun f (map ?I bef @ ?I ?Ds # map ?I aft)) \<in> ?R" (is "(?Il, ?Ir) \<in> _") .         
        from choice have choice: "?cond \<longrightarrow> (?Il, ?Ir) \<in> monoS" using ctxt_closed_one[OF mode_S_ctxt[OF _ mode]] by blast
        from True snbta have left: "?I (Fun f ?bta) = ?Il" by auto
        from step snbta have snbsa: "SN_on (qrstep nfs Q (set R)) {Fun f ?bsa}" by (rule step_preserves_SN_on)
        from IH2 ctxt_closureI[where C = D] have "(?Dt,?Ds) \<in> rstep (set R)"
          unfolding rstep_CC_CS s t by auto
        hence subset: "[tcap (set R) ?Ds] \<subseteq> [tcap (set R) ?Dt]" by (rule tcap_rewrite)		
        have "i_trans_check (Fun f ?bsa)"
        proof (simp add: i_trans_check.simps fus, rule ballI) 
          fix lr'
          assume lr': "lr' \<in> set R - ur"
          show "\<not> (\<lambda>(l, r). Ground_Context.match (GCFun f (map (tcap (set R)) bef @ tcap (set R) ?Ds # map (tcap (set R)) aft)) l) lr'"
          proof (cases lr')
          case (Pair l r)
            from lr' Pair have "(l,r) \<in> set R" by auto
            from Pair lr' have nMatch: 
              "\<not> Ground_Context.match (GCFun f (map (tcap (set R)) bef @ tcap (set R) ?Dt # map (tcap (set R)) aft)) l" (is "\<not> Ground_Context.match ?cbta l") 
              using True by (auto simp: i_trans_check.simps)
            {
              assume "Ground_Context.match (GCFun f (map (tcap (set R)) bef @ tcap (set R) ?Ds # map (tcap (set R)) aft)) l" 
              from this subset have "Ground_Context.match ?cbta l" by (rule match_below)
              with nMatch have False ..
            }
            with lr' Pair show ?thesis by auto 
          qed
        qed
        with snbsa  have "?I (Fun f ?bsa) = ?Ir" by auto	
        with left piece0 choice show ?thesis unfolding s t by (simp)
      qed
    qed
  qed
qed

no_notation Ground_Context.equiv_class ("[_]")
	
context ce_af_redpair
begin

lemma i_transVI: assumes 
  nfs: "nfs = False" and (* required from previous lemmas *)
  mode: "mode_cond mode m" and
  m: "m \<ge> n" and
  sn: "SN_on (qrstep nfs Q (set R)) {t}" and
  step: "(t,s) \<in> (qrstep nfs Q (set R))^*" and
  ur_closed: "ur_closed_af ur us Q (set R) \<pi>" and
  orient: "ur \<subseteq> mode_left mode"
  shows "(itrans.i_trans nfs Q R ur us (c,d,m) t, itrans.i_trans nfs Q R ur us (c,d,m) s) \<in> mode_NS mode"
 (is "?goal t s")
using step 
proof (induct rule: rtrancl_induct)
  case base
  show ?case by (cases mode, unfold mode_NS_def, auto)
next
  case (step s u)
  let ?cn = "(c,d,m)"
  interpret itrans nfs Q R ur us ?cn .
  let ?I = "i_trans"
  from step(1) sn have SN: "SN_on (qrstep nfs Q (set R)) {s}" 
    by (rule steps_preserve_SN_on[of _ _ "qrstep nfs Q (set R)"])
  from step(2)[unfolded qrstep_rule_conv[where R = "set R"]] obtain lr where lr: "lr \<in> set R" and sustep: "(s,u) \<in> qrstep nfs Q {lr}" by auto
  from i_transV[OF nfs mode m lr SN sustep ur_closed orient]
  have isteps2: "(?I s, ?I u) \<in> mode_NS mode" by auto
  with step(3) have "(?I t, ?I u) \<in> mode_NS mode O mode_NS mode" by auto
  thus ?case unfolding mode_NS_def by (cases mode, auto)
qed


lemma i_transVI_S: assumes 
  nfs: "nfs = False" and (* required from previous lemmas *)
  mode: "mode_cond mode m" and
  modeT: "mode" and
  m: "m \<ge> n" and
  sn: "SN_on (qrstep nfs Q (set R)) {t}" and
  step: "(t,s) \<in> (qrstep nfs Q (set R))^* O qrstep nfs Q (set R') O (qrstep nfs Q (set R))^*" and
  ur_closed: "ur_closed_af ur us Q (set R) \<pi>" and
  orient: "ur \<subseteq> mode_left mode" and
  R': "set R' \<inter> ur \<subseteq> Snwf us" and
  RR': "set R' \<subseteq> set R"
  shows "(itrans.i_trans nfs Q R ur us (c,d,m) t, itrans.i_trans nfs Q R ur us (c,d,m) s) \<in> monoS"
proof -
  let ?R = "qrstep nfs Q (set R)"
  let ?R' = "qrstep nfs Q (set R')"
  from step obtain u v where tu: "(t,u) \<in> ?R^*" and uv: "(u,v) \<in> ?R'" and vs: "(v,s) \<in> ?R^*" by auto
  note itVI = i_transVI[OF nfs mode m _ _ ur_closed orient]
  let ?cn = "(c,d,m)"
  let ?I = "itrans.i_trans nfs Q R ur us ?cn"
  from itVI[OF sn tu] have one: "(?I t,?I u) \<in> mode_NS mode" .
  from steps_preserve_SN_on[OF tu sn] have sn: "SN_on ?R {u}" .
  from uv[unfolded qrstep_rule_conv[where R = "set R'"]] obtain lr where lr: "lr \<in> set R'"
    and uv: "(u,v) \<in> qrstep nfs Q {lr}" by auto
  with RR' R' have lr': "lr \<in> set R" and uv': "(u,v) \<in> ?R" and lr: "lr \<notin> ur \<or> lr \<in> Snwf us" 
    unfolding qrstep_rule_conv[where R = "set R"] by auto
  from i_transV[OF nfs mode m lr' sn uv ur_closed orient] modeT lr
  have two: "(?I u,?I v) \<in> monoS" by auto
  from step_preserves_SN_on[OF uv' sn] have sn: "SN_on ?R {v}" .
  from itVI[OF sn vs] have three: "(?I v, ?I s) \<in> mode_NS mode" .
  from one two three have steps: "(?I t, ?I s) \<in> mode_NS mode O monoS O mode_NS mode" (is "_ \<in> ?Rel") by blast
  from modeT have m: "mode_NS mode = monoNS" unfolding mode_NS_def by auto
  have "?Rel \<subseteq> monoS" unfolding m by regexp
  with steps show "(?I t, ?I s) \<in> monoS" by blast
qed
end

abbreviation ur_P_closed_af :: "('f,'v)terms \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f \<times> nat)set \<Rightarrow> ('f,'v)trs \<Rightarrow> 'f af \<Rightarrow> bool" 
  where "ur_P_closed_af Q R ur us P \<pi> \<equiv> \<forall> st \<in> P. ur_closed_term_af (snd st) Q R ur us \<pi>"  

lemma (in ce_af_redpair) i_trans_sound_dp: 
  assumes
  nfs: "nfs \<Longrightarrow> Q \<noteq> {} \<Longrightarrow> wwf_qtrs Q (set R)" and (* required from previous lemmas *)
  mode: "mode_cond mode m"
  and m: "m \<ge> n"
  and P: "(s,t) \<in> P"
  and steps: "(t \<cdot> \<sigma>, u \<cdot> \<tau>) \<in> (qrstep nfs Q (set R))^*"
  and SN: "SN_on (qrstep nfs Q (set R)) {t \<cdot> \<sigma>}"
  and ur_closed: "ur_closed_af ur us Q (set R) \<pi>"
  and ur_P_closed: "ur_P_closed_af Q (set R) ur us P \<pi>"
  and orient: "ur \<subseteq> mode_left mode"
  shows  "(t \<cdot> itrans.i_trans_subst False Q R ur us (c,d,m) \<sigma>, 
    u \<cdot> itrans.i_trans_subst False Q R ur us (c,d,m) \<tau>) \<in> mode_NS mode"
proof -
  let ?cn = "(c,d,m)"
  interpret itrans False Q R ur us ?cn .
  let ?s = "s \<cdot> \<sigma>"
  let ?t = "t \<cdot> \<sigma>"
  let ?u = "u \<cdot> \<tau>"
  let ?I = "i_trans"
  let ?Is = "i_trans_subst"
  let ?ssig = "s \<cdot> ?Is \<sigma>"
  let ?tsig = "t \<cdot> ?Is \<sigma>"
  let ?utau = "u \<cdot> ?Is \<tau>"
  have switch: "qrstep nfs Q (set R) = qrstep False Q (set R)"
    by (rule wwf_qtrs_imp_nfs_False_switch[OF nfs])
  note steps = steps[unfolded switch]
  note SN = SN[unfolded switch]
  from i_transVI[OF refl mode m SN steps ur_closed orient]
  have one: "(?I ?t, ?I ?u) \<in> mode_NS mode" .  
  from P ur_P_closed have "ur_closed_term_af t Q (set R) ur us \<pi>" by auto
  from i_transI[OF this SN] have two: "(?tsig, ?I ?t) \<in> NS^*" by simp
  from steps SN have sns: "SN_on (qrstep False Q (set R)) {?u}" by (rule steps_preserve_SN_on[of _ _ "qrstep False Q (set R)"])
  from i_transII[OF sns, THEN conjunct1] ce_orient[OF m]
  have three: "(?I ?u, ?utau ) \<in> NS^*" by auto    
  from one two three have "(?tsig,?utau) \<in> NS^* O mode_NS mode O NS^*" by blast
  thus ?thesis by simp
qed

lemma (in mono_ce_af_redtriple) mono_redpair_ur_sound:
  assumes mono: "ctxt_closed S"
  and rules: "ur \<subseteq> NS \<union> S"
  and nonStrict: "P \<union> Pw \<subseteq> NS \<union> S"
  and strictP: "Ps - {lr | lr. \<not> funas_term (fst lr) \<subseteq> us} \<subseteq> S"
  and strictR: "(Rs \<inter> ur) - {lr | lr. \<not> funas_term (fst lr) \<subseteq> us} \<subseteq> S"
  and ur_closed: "ur_closed_af ur us Q (set (R @ Rw)) \<pi>"
  and ur_P_closed: "ur_P_closed_af Q (set (R @ Rw)) ur us (P \<union> Pw) \<pi>"
  and m: m
  and nfs: "nfs \<Longrightarrow> Q \<noteq> {} \<Longrightarrow> wwf_qtrs Q (set (R @ Rw))"
  and minChain: "min_ichain (nfs,m,P,Pw,Q,set R,set Rw) s t \<sigma>"
  shows  "\<exists> j. min_ichain (nfs,m,P - Ps, Pw - Ps, Q, set R - Rs, set Rw - Rs) (shift s j) (shift t j) (shift \<sigma> j)"
proof (rule min_ichain_split[OF minChain], rule)
  {
    fix RR
    assume RR: "RR \<subseteq> set (R @ Rw)"
    from RR nfs have "nfs \<Longrightarrow> Q \<noteq> {} \<Longrightarrow> wwf_qtrs Q RR" unfolding wwf_qtrs_def by blast
    from wwf_qtrs_imp_nfs_False_switch[OF this] have "qrstep nfs Q RR = qrstep False Q RR" .
  } note switch_gen = this
  from switch_gen[OF subset_refl] have switch: "qrstep nfs Q (set (R @ Rw)) = qrstep False Q (set (R @ Rw))" .
  assume chain: "min_ichain (nfs,m,Ps \<inter> (P \<union> Pw), P \<union> Pw - Ps, Q, Rs \<inter> (set R \<union> set Rw), set R \<union> set Rw - Rs) s t \<sigma>"
  hence chain: "min_ichain (nfs,m,Ps \<inter> (P \<union> Pw), P \<union> Pw - Ps, Q, Rs \<inter> (set (R @ Rw)), set (R @ Rw) - Rs) s t \<sigma>" by auto    
  have id: "Ps \<inter> (P \<union> Pw) \<union> (P \<union> Pw - Ps) = P \<union> Pw" by auto
  have idR: "Rs \<inter> set (R @ Rw) \<union> (set (R @ Rw) - Rs) = set (R @ Rw)" by auto
  have idS: "S \<union> NS = NS \<union> S" by auto
  from S_ce_compat obtain n' where S_compat: "\<And> m. m \<ge> n' \<Longrightarrow> ce_trs (c,d,m) \<subseteq> S" by blast
  obtain k where k: "k = max n n'" by auto
  hence n: "n \<le> k"  by simp
  obtain mode where mode: "mode" by blast
  from S_compat k mode have cond: "mode_cond mode k" unfolding mode_cond_def using mono by auto
  interpret itrans False Q "(R @ Rw)" ur us "(c, d, k)" .
  obtain \<tau> where \<tau>: "\<tau> = (\<lambda> i. i_trans_subst (\<sigma> i))" by auto
  from mode rules have orient: "ur \<subseteq> mode_left mode" unfolding mode_left_def by auto
  note chain = chain[unfolded min_ichain.simps ichain.simps minimal_cond_def id idR]
  {
    fix i    
    from chain \<open>m\<close> switch
    have P: "(s i, t i) \<in> P \<union> Pw" and R: "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (qrstep False Q (set (R @ Rw)))^*" and 
    SN: "SN_on (qrstep False Q (set (R @ Rw))) {t i \<cdot> \<sigma> i}" by auto
    from i_trans_sound_dp[OF nfs cond n P R SN ur_closed ur_P_closed orient]
    have "(t i \<cdot> \<tau> i, s (Suc i) \<cdot> \<tau> (Suc i)) \<in> (NS \<union> S)^*" unfolding \<tau> mode_NS_def using mode by auto
    hence "(t i \<cdot> \<tau> i, s (Suc i) \<cdot> \<tau> (Suc i)) \<in> (rstep (NS \<union> S))^*" 
      using rtrancl_mono[OF subset_rstep[of "NS \<union> S"]] by auto 
  } note steps = this
  have chain: "ichain (nfs,m,S,NS,{},S,NS) s t \<tau>"
    unfolding ichain.simps qrstep_rstep_conv idS
  proof (intro conjI allI)
    fix i
    from chain nonStrict show "(s i, t i) \<in> NS \<union> S" by auto 
  next
    fix i
    show "(s i \<cdot> \<tau> i) \<in> NF_terms {}" by simp
  next
    fix i
    show "(t i \<cdot> \<tau> i, s (Suc i) \<cdot> \<tau> (Suc i)) \<in> (rstep (NS \<union> S))^*"
      using steps[of i] by simp
  next
    let ?Q = "qrstep nfs Q (set (R @ Rw))"
    let ?QQ = "?Q^* O qrstep nfs Q (Rs \<inter> set (R @ Rw)) O ?Q^*"
    let ?R = "rstep (NS \<union> S)"
    let ?RR = "?R^* O rstep S O ?R^*"
    show "(INFM i. (s i, t i) \<in> S) \<or> 
      (INFM i. (t i \<cdot> \<tau> i, s (Suc i) \<cdot> \<tau> (Suc i)) \<in> ?RR)" unfolding INFM_disj_distrib[symmetric] 
      unfolding INFM_nat
    proof (intro allI)
      fix m
      from chain 
      have "(INFM i. (s i, t i) \<in> Ps \<inter> (P \<union> Pw)) \<or>
        (INFM i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?QQ)" by blast
      from this[unfolded INFM_disj_distrib[symmetric], unfolded INFM_nat]
      obtain l where l: "Suc m < l" and alt:"(s l, t l) \<in> Ps \<inter> (P \<union> Pw) \<or> 
        (t l \<cdot> \<sigma> l, s (Suc l) \<cdot> \<sigma> (Suc l)) \<in> ?QQ" by blast
      hence l': "m < l" by auto
      let ?I = "i_trans"
      obtain I where I: "?I = I" by auto        
      from alt show "\<exists> l > m. (s l, t l) \<in> S \<or> (t l \<cdot> \<tau> l, s (Suc l) \<cdot> \<tau> (Suc l)) \<in> ?RR"
      proof
        assume Ssteps: "(t l \<cdot> \<sigma> l, s (Suc l) \<cdot> \<sigma> (Suc l)) \<in> ?QQ"
        let ?ts = "t l \<cdot> \<sigma> l"
        let ?tt = "t l \<cdot> \<tau> l"
        let ?ss = "s (Suc l) \<cdot> \<sigma> (Suc l)"      
        let ?st = "s (Suc l) \<cdot> \<tau> (Suc l)"              
        from chain m switch have SN: "SN_on (qrstep False Q (set (R @ Rw))) {?ts}" by simp
        from chain have P: "(s l, t l) \<in> P \<union> Pw" by auto
        from P ur_P_closed have urt: "ur_closed_term_af (t l) Q (set (R @ Rw)) ur us \<pi>" by force
        from chain have steps: "(?ts, ?ss) \<in> ?Q^*" by auto
        from steps[unfolded switch] SN have SNs: "SN_on (qrstep False Q (set (R @ Rw))) {?ss}" 
          by (rule steps_preserve_SN_on[of _ _ "qrstep False Q (set (R @ Rw))"])
        from i_transI[OF urt SN] have ns0: "(?tt, ?I ?ts) \<in> NS^*" unfolding \<tau> .
        from i_transII[OF SNs, THEN conjunct1] ce_orient[OF n]
        have ns2: "(?I ?ss, ?st) \<in> NS^*" unfolding \<tau> by auto
        let ?Rs = "filter (\<lambda> lr. lr \<in> Rs) (R @ Rw)"
        have Rs: "Rs \<inter> (set (R @ Rw)) = set ?Rs" by auto 
        from switch_gen[of "set ?Rs"] have switch2: "qrstep nfs Q (set ?Rs) = qrstep False Q (set ?Rs)" by auto
        note Ssteps = Ssteps[unfolded Rs switch switch2]        
        have "(?I ?ts, ?I ?ss) \<in> monoNS O S O monoNS"
          by (rule i_transVI_S[OF refl cond mode n SN Ssteps ur_closed orient], insert strictR, auto)
        from ns0 this ns2 have dec: "(?tt, ?st) \<in> NS^* O monoNS O S O monoNS O NS^*" unfolding I by auto
        have "NS^* O monoNS O S O monoNS O NS^* \<subseteq> monoNS O S O monoNS" by regexp
        with dec have "(?tt, ?st) \<in> monoNS O S O monoNS" by blast
        with rtrancl_mono[OF subset_rstep[of "NS \<union> S"]] subset_rstep[of S] have "(?tt, ?st) \<in> ?RR" by auto
        with l' show ?thesis by auto
      next
        assume P: "(s l, t l) \<in> Ps \<inter> (P \<union> Pw)"
        with strictP have "(s l, t l) \<in> S \<or> \<not> funas_term (s l) \<subseteq> us" by force
        thus ?thesis
        proof
          assume "(s l, t l) \<in> S" thus ?thesis using l' by auto
        next
          assume nwf: "\<not> funas_term (s l) \<subseteq> us"
          from l obtain ll where ll: "l = Suc ll" by (cases l, auto)
          let ?ts = "t ll \<cdot> \<sigma> ll"
          let ?tt = "t ll \<cdot> \<tau> ll"
          let ?ss = "s (Suc ll) \<cdot> \<sigma> (Suc ll)"      
          let ?st = "s (Suc ll) \<cdot> \<tau> (Suc ll)"              
          from chain m switch have SN: "SN_on (qrstep False Q (set (R @ Rw))) {?ts}" by simp
          from chain have P: "(s ll, t ll) \<in> P \<union> Pw" by auto
          from P ur_P_closed have urt: "ur_closed_term_af (t ll) Q (set (R @ Rw)) ur us \<pi>" by force
          from chain have steps: "(?ts, ?ss) \<in> ?Q^*" by auto
          note steps = steps[unfolded switch]
          from steps SN have SNs: "SN_on (qrstep False Q (set (R @ Rw))) {?ss}" by (rule steps_preserve_SN_on)
          from i_transI[OF urt SN] have ns0: "(?tt, ?I ?ts) \<in> NS^*" unfolding \<tau> .
          from i_transVI[OF refl cond n SN steps ur_closed orient] have ns1: "(?I ?ts, ?I ?ss) \<in> monoNS"
            using mode unfolding mode_NS_def by auto
          from i_transII[OF SNs, THEN conjunct2] nwf[unfolded ll]
          have "(?I ?ss, ?st) \<in> (rstep (ce_trs (c,d,k)))^+" unfolding \<tau> by auto
          from trancl_mono[OF this mono_ce[OF cond[unfolded mode_cond_def, THEN mp[OF _ mode]]]]
          have strict: "(?I ?ss, ?st) \<in> S^+" .
          from ns0 ns1 strict have strict: "(?tt,?st) \<in> NS^* O monoNS O S^+" by auto
          have "NS^*O monoNS O S^+ \<subseteq> monoNS O S O monoNS" by regexp
          from set_mp[OF this strict] have "(?tt,?st) \<in> monoNS O S O monoNS" .
          with rtrancl_mono[OF subset_rstep[of "NS \<union> S"]] subset_rstep[of S] have strict: "(?tt, ?st) \<in> ?RR" by auto
          from l ll have "m < ll" by auto
          with strict show ?thesis by blast
        qed
      qed
    qed
  next
    fix i
    show "NF_subst nfs (s i, t i) (\<tau> i) {}" by simp
  qed
  have idS: "S \<inter> (NS \<union> S) = S" by auto
  from ichain_mono[OF chain subset_refl _ subset_refl subset_refl, of "NS \<union> S - S" "NS \<union> S - S"] 
    mono_redpair_sound_ichain[OF mono subset_refl subset_refl, of nfs m "{}" s t \<tau>]
  show False unfolding idS by blast
qed

end